#!/usr/bin/env python
# -*- coding: utf-8 -*-

# # Leçon de code
# 
# La programmation est devenue un outil essentiel du datascientist
# mais pas seulement. Beaucoup d'outils pointus sont open source
# mais uniquement accessibles à ceux qui savent programmer.
# Après la mise au point d'un modèle statistique, économique,
# il se pose souvent la question de la mise à jour fréquente des
# résultats, c'est à dire leur automatisation via la programmation.

# In[1]:


from jyquickhelper import add_notebook_menu
add_notebook_menu()


# ## Après la prépas
# 
# Programmation, ENSAE, automatisation, emplois

# ### ENSAE
# 
# * statistique, finance, économie, actuariat, data science
# * utilisation massive des données
# * plus de logiciel passe partout

# ### Excel, Matlab, SAS, Python, Notebook
# 
# * tableur : difficile de passer à l'échelle
# * SAS, Matlab : onéreux
# * Python : open source, langage accessible sans être un expert, effort de design comme avec [scikit-learn](https://scikit-learn.org/stable/) ([INRIA](https://scikit-learn.fondation-inria.fr/people/))
# * [notebook](https://jupyter.org/) : outils permettant de mélanger texte, formules, code et de le partager

# ### Python, R : Python
# 
# * tous deux open sources
# * python plus complet
# * attire les développeurs, pas seulement les chercheurs
# * R plus proche de matlab

# ### Installation
# 
# * [Anaconda](https://anaconda.org/)

# ### A distance
# 
# * [colab](https://colab.research.google.com/)

# ### Après la prépa
# 
# * code efficace
# * test unitaires, exceptions $\longrightarrow$ à connaître pour un entretien d'embauche
# * packaging : exemple avec [2048](http://www.xavierdupre.fr/app/pystrat2048/helpsphinx/index.html)
# * manipulation de données, dataframes, graphes.

# ### Un exemple : coût de l'algorithme ?

# In[2]:


def position_max(tableau):
    for i in range(0, len(tableau)):
        if tableau[i] == max(tableau):
            return i

mx = position_max([6, 7, 4, 11, -5, 4])
mx


# ### Ressource
# 
# * contenus en ligne [xavierdupre.fr](http://www.xavierdupre.fr/)
# * Google, Bing, Duck Duck Go, Qwant, Baidu, Yandex
# * [Stackoverflow](https://stackoverflow.com/)
# * [openclassroom](https://openclassrooms.com/fr/)
# * mail au professeur

# ### COVID
# 
# * apprentissage pour tout le monde y compris pour les encadrants
# * vidéo, pédagogie inversée
# * projet en groupe : application Flask

# ## Cours

# ### Objectif du cours
# 
# * [Test unitaire](http://www.xavierdupre.fr/app/ensae_teaching_cs/helpsphinx/notebooks/td1a_unit_test_ci.html)
# * Calcul matriciel avec [numpy](https://numpy.org/)
# * [Culture algorithmique](http://www.xavierdupre.fr/app/ensae_teaching_cs/helpsphinx/specials/algorithm_culture.html)
# * Etre capable de réaliser une application [Flask](https://flask.palletsprojects.com/en/1.1.x/) (web) qui récupère des données pour faire des statistiques.
# * Etre capable de reproduire ou réutiliser l'algorithme décrit dans un article ou implémenté sur github

# ### Organisation
# 
# * Répartition en TDs
# * Le chemin vers les objectifs du cours sont différents selon les TDs

# ## Quizz 1

# ### affectation

# In[3]:


a = 3


# ### types

# In[4]:


a = 3
p = 4.56
b = 'r'
c = (4, 6)
g = [5, 4]
d = {'a': 0, 'b': 1}


# ### test

# In[5]:


h = 7
if h % 2 == 0:
    msg = 'pair'
else:
    msg = 'impair'
print(msg)


# ### boucle

# In[6]:


for element in [4, 5, 8]:
    print(element)


# In[7]:


it = 0
while it < 24:
    print(it)
    it += 5


# ### fonction

# In[8]:


def area(l, w):
    return l * w

print(area(4, 5))


# ### print / return ?

# In[9]:


def area(l, w):
    print(l * w)

print(area(4, 5))


# ### import

# In[10]:


import math
from math import cos
cos(5) + math.sin(5)


# ### classes ?

# In[11]:


class Vase:
    def __init__(self, hauteur, diametre):
        self.hauteur = hauteur
        self.diametre = diametre
    def area(self):
        return self.hauteur * self.diametre * math.pi
    
v = Vase(5, 3)
print(v.area())


# ## Quizz 2 : array, dataframe, graphe
# 
# On ne code plus le produit matriciel. Il est très difficile d'être plus rapide que [numpy](https://numpy.org/) qui utilise des libraires telles que [BLAS](http://www.netlib.org/blas/) qui savent tirer parti des optimisations processeurs [AVX](https://en.wikipedia.org/wiki/Advanced_Vector_Extensions), voire de processeurs différents [GPU](https://en.wikipedia.org/wiki/Graphics_processing_unit).

# ### Produit matriciel

# In[12]:


import numpy
mat1 = numpy.array([[1, 2, 3], [4, 5, 6]])
mat2 = numpy.array([[1, -1], [-1, 1], [0, 0]])
mat1 @ mat2


# In[13]:


mat2 @ mat1


# ### Dataframe

# In[14]:


import pandas
df = pandas.DataFrame([{'col1': 4.5, 'col2': "legend"},
                       {'col1': 4.5, 'col3': -8},
                       {'col1': 14.5, 'col3': -80, 'col2': 'note'}])
df


# In[15]:


df.isna()


# In[16]:


df.isna().astype(numpy.int64)


# In[17]:


df[df['col1'] >= 10]


# ### Graphes

# In[18]:


rnd = numpy.random.randn(50, 3) @ numpy.array([[1, 0, 1], [0, 1, 0], [0, -3, 1]])
rnd[:5]


# In[19]:


get_ipython().run_line_magic('matplotlib', 'inline')


# In[20]:


import matplotlib.pyplot as plt
fig, ax = plt.subplots(1, 3, figsize=(12, 4), sharex=True, sharey=True)
ax[0].plot(rnd[:, 0], rnd[:, 1], '.')
ax[0].set_title("Axes 1 et 2")

ax[1].plot(rnd[:, 1], rnd[:, 2], '.')
ax[1].set_title("Axes 2 et 3")

ax[2].plot(rnd[:, 0], rnd[:, 1], '.', label="Axes 1 et 2")
ax[2].plot(rnd[:, 1], rnd[:, 2], '.', label="Axes 2 et 3")
ax[2].set_title("Axes 1 et 2 et 3")
ax[2].legend();


# ## Quizz 3 : algorithme

# In[21]:


from IPython.display import SVG, Image


# ### Recherche dichotomique

# In[22]:


Image("https://upload.wikimedia.org/wikipedia/commons/f/f7/Binary_search_into_array.png")


# ### Tri fusion

# In[23]:


Image("https://upload.wikimedia.org/wikipedia/commons/6/60/Mergesort_algorithm_diagram.png")


# ### Tas

# In[24]:


SVG("https://upload.wikimedia.org/wikipedia/commons/3/38/Max-Heap.svg")


# ### Plus court chemin dans un graphe

# In[25]:


SVG("https://upload.wikimedia.org/wikipedia/commons/2/29/DijkstraBis01.svg")


# ### Voygeur de commerce

# In[26]:


Image("tsp.png", width=400)


# ### distance d'édition

# In[27]:


Image("https://upload.wikimedia.org/wikipedia/commons/d/d1/Levenshtein_distance_animation.gif")


# ### Problème du sac-à-dos

# In[28]:


SVG("https://upload.wikimedia.org/wikipedia/commons/f/fd/Knapsack.svg")


# ### Simplexe

# In[29]:


Image('https://upload.wikimedia.org/wikipedia/commons/2/25/Tetrahedron.png', width=400)


# ### Postier chinois

# In[30]:


Image("postier.png")


# ### Arbre de décision

# In[31]:


Image("dectree.png")


# ## Quizz 4 : génie logiciel

# ### Les exceptions
# 
# Les programmes qui plantent mais en fait c'est pas grave.

# In[32]:


try:
    y = 1 / 0
except Exception as e:
    print(type(e), e)


# Ou des fois-ci

# In[33]:


try:
    x = -1
    if x < 0:
        raise ValueError("La racine carrée d'un nombre positif est inconnue de ce programme.")
except Exception as e:
    print(type(e), e)


# ### Les expressions régulières

# In[34]:


import re
reg = re.compile("[A-Z]{2,}")
texte = "Etrange ces acronymes comme ENSAE ou CPU qui ressortent comme par magie."
reg.findall(texte)


# ### Les tests unitaires
# 
# Ou comment protéger son code contre l'intrusion d'un codeur distrait.

# In[35]:


def return_sept(n):
    return int('7' * n)

def test_unitaire():
    # Si ça plante, c'est de votre faute.
    assert return_sept(4) == 7777

test_unitaire()


# ### Le packaging...
# 
# Il y a plusieurs façons de passer à la postérité. C'est l'une d'elle.

# ## Problèmes, Exercices

# ### Enumérer les permutations
# 
# [enumerate_permutations_recursive](http://www.xavierdupre.fr/app/ensae_teaching_cs/helpsphinx/ensae_teaching_cs/td_1a/construction_classique.html?highlight=classique#ensae_teaching_cs.td_1a.construction_classique.enumerate_permutations_recursive).

# ### Suggestions

# In[36]:


Image("suggestion.png")


# Le **trie**

# In[37]:


SVG('https://upload.wikimedia.org/wikipedia/commons/b/be/Trie_example.svg')


# ### Récupérer des mails automatiquement
# 
# ...

# ### Calculer des statistiques et les envoyer automatiquement au format PDF par mail
# 
# ...

# ### Ecrire un système qui note automatiquement les présences
# 
# par reconnaissance faciale mais qui échoue pour cause de masques qui passe alors par la voix en demandant de chanter du Johnny.

# ## A suivre.

# In[38]: